from __future__ import annotations

from typing import Dict
from typing import Final
from typing import Literal
from typing import NamedTuple
from typing import Tuple
from typing import Type
from typing import Union

from ap_games.player.player import Player

EMPTY: Final[Literal[' ']] = ' '
O_MARK: Final[Literal['O']] = 'O'
X_MARK: Final[Literal['X']] = 'X'
UNDEFINED_MARK: Final[Literal['']] = ''

PlayerMark = Literal['X', 'O']
Mark = Union[Literal['X'], Literal['O'], Literal[' '], Literal['']]

SupportedPlayers = Dict[str, Type[Player]]
OptionalPlayerTypes = Union[Tuple[str, str], Tuple[()]]
Size = int


class Coordinate(NamedTuple):
    """Coordinate(x: int, y: int)."""

    x: int  # noqa: WPS111
    y: int  # noqa: WPS111


class Cell(NamedTuple):
    """Cell(coordinate: Coordinate, mark: str)."""

    coordinate: Coordinate
    mark: Mark


class Offset(NamedTuple):
    """Offset(coordinate: Coordinate, direction: Coordinate)."""

    coordinate: Coordinate
    direction: Coordinate


class GameStatus(NamedTuple):
    """GameStatus(active: bool, message: str, must_skip: bool)."""

    active: bool
    message: str
    must_skip: bool


class Move(NamedTuple):
    """Move(coordinate: Coordinate, score: int, percentage: int, last: bool).

    :ivar coordinate:  The coordinate of selected cell or
        ``undefined_coordinate`` if game status is ``False``.
    :ivar score:  The terminal game score.
    :ivar percentage:  The percentage to reach ``score`` as a number
        greater 0 and less than or equal to 100.  See description above.
    :ivar last:  ``True`` if current move finishes the game, ``False`` -
        otherwise.

    """

    coordinate: Coordinate
    score: int
    percentage: int
    last: bool


class Node(NamedTuple):
    """Move(player_mark: PlayerMark, move: Move, sub_tree: Tree)."""

    player_mark: PlayerMark
    move: Move
    sub_tree: 'Tree'  # type: ignore


UNDEFINED_MOVE = Move(Coordinate(x=0, y=0), score=0, percentage=0, last=False)

Side = Tuple[Cell, ...]
Directions = Tuple[Coordinate, ...]
Coordinates = Tuple[Coordinate, ...]
Tree = Dict[str, Node]  # type: ignore



from __future__ import annotations

import logging
from operator import add
from operator import sub
import random
from typing import TYPE_CHECKING

from ap_games.ap_types import Move
from ap_games.ap_types import Node
from ap_games.ap_types import UNDEFINED_MOVE
from ap_games.gameboard.gameboard import SquareGameboard
from ap_games.log import logger
from ap_games.player.player import Player

if TYPE_CHECKING:
    from typing import ClassVar
    from typing import Dict
    from typing import List
    from typing import Optional
    from typing import Set

    from ap_games.ap_types import Coordinate
    from ap_games.ap_types import GameStatus
    from ap_games.ap_types import PlayerMark
    from ap_games.ap_types import Tree
    from ap_games.game.game_base import TwoPlayerBoardGame

__all__ = ('AIPlayer',)


class AIPlayer(Player):
    """AIPlayer in the game."""

    _max_depth: ClassVar[Dict[str, int]] = {
        'easy': 0,
        'medium': 2,
        'hard': 4,
        'nightmare': 6,
    }

    def __init__(
        self, type_: str, /, *, mark: PlayerMark, game: TwoPlayerBoardGame
    ) -> None:
        super().__init__(type_, mark=mark, game=game)
        self.max_depth = self._max_depth[type_]
        self.tree: Tree = {}

    def move(self) -> Coordinate:
        """Define coordinate of the next move and return it.

        :returns: Coordinate chosen according to the minimax algorithm
            when :attr:`.max_depth` is not equal to 0.

        """
        logger.info(f'Making move level "{self.type_}" [{self.mark}]')

        if self.max_depth:
            return self._minimax().coordinate
        return self._random_coordinate()

    def _minimax(  # noqa: C901
        self,
        gameboard: Optional[SquareGameboard] = None,
        player_mark: Optional[PlayerMark] = None,
        depth: int = 0,
        tree: Optional[Tree] = None,
    ) -> Move:
        """Return the move selected by the minimax algorithm.

        Mini-max algorithm:

            1. Return the selected move with the terminal score, if a
               terminal state is achieved;
            2. Go through available moves on the board
               (:meth:`._go_through_available_moves`) or go through
               cached moves in :attr:`.tree`
               (:meth:`._go_through_subtree`). See docstrings
               corresponding method for details;
            3. Call the :meth:`._minimax` method on each available move
               (recursion) (:meth:`._get_terminal_score`);
            4. Evaluate returning values from minimax-method calls
               (:meth:`._choose_best_move`).  See next methods for
               details:

               * :meth:`._correct_priority_coordinates`;
               * :meth:`._extract_desired_moves`;
               * :meth:`._extract_most_likely_moves`.

            5. Return the best ``Move``.

        :param gameboard:  Optional.  If undefined, use
            :attr:`.game.gameboard`.  The gameboard relative to which
            the terminal score of the game will be calculated.
        :param player_mark:  Optional.  If undefined, use ``self.mark``.
            The player relative to whom the terminal score of the game
            will be calculated.
        :param depth:  Optional.  ``0`` by default.  The current depth
            of tree.
        :param tree:  Optional.  If undefined, use ``self.tree``.  It is
            cached tree possible cases as ``dict``.  Where key is a grid
            of the gameboard (as string), value is a nemedtuple ``Node``
            three fields.

            .. seealso::

                :class:`Node`.

        ``Percentage``::

            .. note::

                This is only important when the 'depth' of analysis is
                limited.

            In the minimax algorithm, it doesn't matter how many ways
            to win AI at the end of the game. Therefore, the AI
            'stops fighting' and is not trying to 'steal' one of them.
            With the variable ``percentage``, the case with two
            possible moves to lose are worse than case with one.

            Run example below with and without variable ``percentage``
            once or twice::

                >>> from ap_games.game.tictactoe import TicTacToe
                >>> TicTacToe(
                ...     grid='X_OXX_O__',
                ...     player_types=('easy', 'hard')
                ... ).play()

            .. note::

                "hard" select cell randomly from all empty cells and can
                lose to "easy" without ``percentage``.

        ``Factor``::

            .. note::

                This is only important when the 'depth' of analysis is
                limited.

            In the minimax algorithm, it doesn't matter when you lose:
            now or later. Therefore, the AI 'stops fighting' if it
            in any case loses the next moves, regardless of how it takes
            the move now. In this case, the AI considers that all the
            moves are the same bad.

            Because the enemy can make a mistake, and adding the
            variable ``last_move_coefficient`` allows the AI to use a
            possible enemy errors in the future.

            With the ``last_move_coefficient``, losing now is worse than
            losing later.  Therefore, the AI is trying not to 'give up'
            now and wait for better chances in the future.

            Run example below with and without variable
            ``last_move_coefficient`` once or twice:

                >>> TicTacToe(
                ...     grid='X_OX_____',
                ...     player_types=('easy', 'hard')
                ... ).play()

            .. note::

                'hard' select cell randomly from all empty cells and
                can lose to 'easy' without ``last_move_coefficient``.

        :returns:  The move is selected according to the minimax
            algorithm as a namedtuple :class:`Move` instance.

            .. seealso::

                :class:`Move`.

        """
        if gameboard is None:
            gameboard = self.game.gameboard
        if player_mark is None:
            player_mark = self.mark
        if tree is None:
            self._unpack_tree()
            tree = self.tree

        node: Node = tree.setdefault(
            gameboard.grid_as_string,
            Node(player_mark=player_mark, move=UNDEFINED_MOVE, sub_tree={},),
        )
        last_move_coefficient: int = 1
        last: bool = False

        game_status: GameStatus = self.game.get_status(gameboard, player_mark)

        if game_status.must_skip:
            player_mark = self.game.get_enemy_mark(player_mark)
            depth -= 1
            game_status = game_status._replace(active=True)

        if game_status.active:
            if depth < self.max_depth:
                if node.sub_tree:
                    return self._go_through_subtree(
                        depth=depth + 1, tree=node.sub_tree,
                    )
                else:  # node.sub_tree == {}
                    return self._go_through_available_moves(
                        gameboard=gameboard,
                        player_mark=player_mark,
                        depth=depth + 1,
                        tree=node.sub_tree,
                    )
        else:
            # 10 >= max possible self.game.gameboard.size
            last_move_coefficient = 10 * 10 ** (self.max_depth - depth + 1)
            last = True

        # in minimax algorithm ``score`` is always computed relative to
        # current (``self``) player
        score: int = self.game.get_score(gameboard, player_mark=self.mark)
        return Move(
            coordinate=self.game.gameboard.undefined_coordinate,
            score=score * last_move_coefficient,
            percentage=100,
            last=last,
        )

    def _go_through_available_moves(
        self,
        gameboard: SquareGameboard,
        player_mark: PlayerMark,
        depth: int,
        tree: Tree,
    ) -> Move:
        """Call minimax method on each available move.

        :param gameboard:  The gameboard relative to which the terminal
            score of the game will be calculated.
        :param player_mark:  The mark of player relative to whom the
            terminal score of the game will be calculated.
        :param depth:  The current depth of tree.
        :param tree:  The cache that will be filled up by this method.

            .. warning::

                ``tree`` is always an empty dict.

        :returns:  The move selected by the minimax algorithm as
            instance of namedtuple :class:`Move`.

        """
        moves: List[Move] = []
        for coordinate in self.game.get_available_moves(
            gameboard, player_mark
        ):
            fake_gameboard: SquareGameboard = gameboard.copy(
                indent='\t' * depth
            )
            self.game.place_mark(coordinate, player_mark, fake_gameboard)

            move: Move = self._get_terminal_score(
                coordinate=coordinate,
                player_mark=player_mark,
                gameboard=fake_gameboard,
                depth=depth,
                tree=tree,
            )
            moves.append(move)
        return self._choose_best_move(moves, player_mark, depth)

    def _go_through_subtree(self, depth: int, tree: Tree) -> Move:
        """Call minimax method on each node of tree.

        :param depth:  The current depth of tree.
        :param tree:  The tree of cached moves.

            .. warning::  ``tree`` is always not empty dict.

        :raises ValueError: When there are nodes in the same level with
            different ``player_mark``.

        :returns:  The move selected by the minimax algorithm as
            instance of namedtuple :class:`Move`.

        """
        player_marks: Set[PlayerMark] = set()
        moves: List[Move] = []
        for grid, node in tree.items():
            if node.move.last:
                moves.append(node.move)
            else:
                coordinate: Coordinate = node.move.coordinate
                player_mark: PlayerMark = node.player_mark
                indent: str = '\t' * depth
                if node.sub_tree:
                    if logger.level == logging.DEBUG:
                        logger.debug(f'\n{indent}[{player_mark}] {coordinate}')
                        logger.debug(f'{indent}[{grid}]')
                    move: Move = self._go_through_subtree(
                        depth=depth + 1, tree=node.sub_tree
                    )
                    move = move._replace(coordinate=coordinate)
                    tree[grid] = Node(
                        player_mark=player_mark,
                        move=move,
                        sub_tree=node.sub_tree,
                    )
                    moves.append(move)
                else:  # node.sub_tree == {}:
                    # don't compare ``depth`` and ``max_depth``, because
                    # ``max_depth`` doesn't change during game.
                    # Therefore cached ``tree`` will never be deeper
                    # than ``max_depth``.
                    fake_gameboard: SquareGameboard = SquareGameboard(
                        grid=grid, indent=indent, colorized=False,
                    )

                    move = self._get_terminal_score(
                        coordinate=coordinate,
                        player_mark=player_mark,
                        gameboard=fake_gameboard,
                        depth=depth,
                        tree=tree,
                    )
                    moves.append(move)
            player_marks.add(node.player_mark)
        if len(player_marks) > 1:
            raise ValueError("Impossible!")
        return self._choose_best_move(moves, player_marks.pop(), depth)

    def _get_terminal_score(
        self,
        coordinate: Coordinate,
        player_mark: PlayerMark,
        gameboard: SquareGameboard,
        depth: int,
        tree: Tree,
    ) -> Move:
        if logger.level == logging.DEBUG:
            indent: str = '\t' * depth
            logger.debug(f'\n{indent}[{player_mark}] {coordinate}')
            logger.debug(gameboard)

        next_player_mark: PlayerMark = self.game.get_enemy_mark(player_mark)
        move: Move = self._minimax(
            gameboard=gameboard,
            player_mark=next_player_mark,
            depth=depth,
            tree=tree,
        )
        grid: str = gameboard.grid_as_string
        move = move._replace(coordinate=coordinate)
        sub_tree: Tree = tree[grid].sub_tree
        tree[grid] = Node(
            player_mark=player_mark, move=move, sub_tree=sub_tree
        )
        return move

    def _unpack_tree(self) -> None:
        node: Node
        grid: str = self.game.gameboard.grid_as_string
        while self.tree and (grid not in self.tree):
            new_tree: Tree = {}
            for node in self.tree.values():
                if node.sub_tree:
                    new_tree.update(node.sub_tree)
            self.tree = new_tree

    def _choose_best_move(
        self, moves: List[Move], player_mark: PlayerMark, depth: int,
    ) -> Move:
        indent: str = '\t' * depth

        if logger.level == logging.DEBUG:
            logger.debug(
                f'{indent}Choose the best move from moves -> ' f'{str(moves)}'
            )

        corrected_moves: List[Move] = self._correct_priority_coordinates(
            moves=moves, player_mark=player_mark, depth=depth
        )

        desired_moves: List[Move] = self._extract_desired_moves(
            moves=corrected_moves, player_mark=player_mark, depth=depth
        )

        most_likely_moves: List[Move] = self._extract_most_likely_moves(
            moves=desired_moves, player_mark=player_mark, depth=depth
        )

        move: Move = random.choice(most_likely_moves)
        # compute and replace ``percentage`` in the selected move
        move = move._replace(
            percentage=int(len(desired_moves) / len(moves) * move.percentage)
        )

        if logger.level == logging.DEBUG:
            logger.debug(f'{indent}Selected move: {move}')

        return move

    def _correct_priority_coordinates(
        self, moves: List[Move], player_mark: PlayerMark, depth: int,
    ) -> List[Move]:
        """Change score of moves from with high priority coordinates.

        .. note::

            This function only makes sense if the minimax algorithm is
            limited in depth and cannot reach the end of the game.

        Function increases "score" of move if it is the move of the
        current player , and decrease "score" of move if it is move of
        enemy player.

        :param moves:  Possible moves that should be checked.
        :param player_mark:  The mark of player who moves.
        :param depth:  Current depth of tree.

        :return:  The list of input ``moves`` with changed score of moves
            whose coordinates are in :attr:`.priority_coordinates`.

        """
        if not self.game.priority_coordinates:
            return moves

        if player_mark == self.mark:
            op = add
        else:
            op = sub
        corrected_moves: List[Move] = []
        for move in moves:
            if move.coordinate in self.game.priority_coordinates:
                corrected_moves.append(
                    move._replace(
                        score=op(
                            move.score,
                            self.game.priority_coordinates[move.coordinate],
                        )
                    )
                )
            else:
                corrected_moves.append(move)
        return corrected_moves

    def _extract_desired_moves(
        self, moves: List[Move], player_mark: PlayerMark, depth: int
    ) -> List[Move]:
        """Calculate min-max score and returning moves with that score.

        Maximize score of self own move or minimize score of enemy
        moves.

        :param moves:  Possible moves that should be checked.
        :param player_mark:  The mark of player who moves.
        :param depth:  Current depth of tree.

        :return:  A new list of moves that is a subset of the input
            moves.

        """
        if player_mark == self.mark:
            score_func = max
        else:
            score_func = min

        desired_score: int = score_func(move.score for move in moves)
        desired_moves: List[Move] = [
            move for move in moves if move.score == desired_score
        ]
        if logger.level == logging.DEBUG:
            indent: str = '\t' * depth
            logger.debug(
                f'{indent}Desired score moves ({score_func}) -> '
                f'{desired_moves}'
            )
        return desired_moves

    def _extract_most_likely_moves(
        self, moves: List[Move], player_mark: PlayerMark, depth: int
    ) -> List[Move]:
        """Maximize probability of self own winning or enemy losing.

        .. warning::

            All input moves on this stage must have the same score.

        :param depth:  Current depth of tree.
        :param moves:  Possible moves that should be checked.
        :param player_mark:  The mark of player who moves and relative
            to which ``percentage_func`` will be determined.

        :return:  A new list of moves that is a subset of the input
            moves.

        """
        desired_score: int = moves[0].score

        if (desired_score > 0 and player_mark == self.mark) or (
            desired_score <= 0 and player_mark != self.mark
        ):
            percentage_func = max
        else:
            percentage_func = min
        desired_percentage: int = percentage_func(
            move.percentage for move in moves
        )
        most_likely_moves: List[Move] = [
            move for move in moves if move.percentage == desired_percentage
        ]
        if logger.level == logging.DEBUG:
            indent: str = '\t' * depth
            logger.debug(
                f'{indent}Desired percentage moves ({percentage_func}) -> '
                f'{str(most_likely_moves)}'
            )
        return most_likely_moves



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

from ap_games.ap_types import Coordinate
from ap_games.ap_types import EMPTY
from ap_games.ap_types import GameStatus
from ap_games.game.game_base import TwoPlayerBoardGame
from ap_games.log import logger

if TYPE_CHECKING:
    from typing import ClassVar
    from typing import Dict
    from typing import List
    from typing import Optional
    from typing import Tuple

    from ap_games.ap_types import Coordinates
    from ap_games.ap_types import Directions
    from ap_games.ap_types import PlayerMark
    from ap_games.gameboard.gameboard import SquareGameboard

__all__ = ('Reversi',)


class Reversi(TwoPlayerBoardGame):
    """Reversi game supports human user and some types of AI.

    For details see :class:`TwoPlayerBoardGame`.

    """

    default_grid: ClassVar[str] = f'{EMPTY * 27}XO{EMPTY * 6}OX{EMPTY * 27}'
    # coordinates with additional score
    priority_coordinates: ClassVar[Dict[Coordinate, int]] = {
        Coordinate(1, 1): 10,
        Coordinate(1, 8): 10,
        Coordinate(8, 8): 10,
        Coordinate(8, 1): 10,
    }

    rules: ClassVar[str] = ''.join(
        (
            "You must place the piece so that an opponent's piece, or a ",
            "row of opponent's pieces, is flanked by your pieces.\nAll of ",
            "the opponent's pieces between your pieces are then turned ",
            'over to become your color. The object of the game is to own ',
            'more pieces than your opponent when the game is over.',
        )
    )

    def get_status(
        self,
        gameboard: Optional[SquareGameboard] = None,
        player_mark: Optional[PlayerMark] = None,
    ) -> GameStatus:
        """Return game status calculated in accordance with game rules.

        .. seealso::

            :meth:`.TwoPlayerBoardGame.get_status`

        :param gameboard: Optional.  If undefined, use
            :attr:`.gameboard`.
        :param player_mark: Optional.  If undefined, use mark of player
            with index ``0`` in :attr:`.players` (current player mark).

        :returns: Game status as the instance of namedtuple
            :class:`GameStatus`.
            ``GameStatus.active == False`` if game cannot be continued.
            ``GameStatus.must_skip == True`` means that the game can
            only be continued after the rotation of the players.

        """
        if gameboard is None:
            gameboard = self.gameboard
        if player_mark is None:
            player_mark = self.players[0].mark

        if self.get_available_moves(gameboard, player_mark):
            return GameStatus(active=True, message='', must_skip=False)

        enemy_mark: PlayerMark = self.get_enemy_mark(player_mark)
        if self.get_available_moves(gameboard, enemy_mark):
            game_status = GameStatus(
                active=False,
                message=(
                    f'\nThe player [{player_mark}] has no moves available!\n'
                ),
                must_skip=True,
            )
        else:
            winners: Tuple[PlayerMark, ...] = self._get_winners(
                gameboard=gameboard
            )
            if len(winners) == 1:
                game_status = GameStatus(
                    active=False,
                    message=f'{winners[0]} wins\n',
                    must_skip=False,
                )
            else:  # len(winners) > 1:
                game_status = GameStatus(
                    active=False, message='Draw\n', must_skip=False
                )
        return game_status

    def place_mark(
        self,
        coordinate: Coordinate,
        player_mark: Optional[PlayerMark] = None,
        gameboard: Optional[SquareGameboard] = None,
    ) -> int:
        """Player's move at given coordinate on the gameboard.

        :param coordinate: The coordinate of cell where player want to
            move.
        :param player_mark: Optional.  If undefined, use mark of current
            player (mark of the player with index ``0`` in
            :attr:`.players`).
        :param gameboard: Optional.  If undefined, use
            :attr:`.gameboard`.

        :returns: Score as count of marked cells.  Return ``0`` if no
            cells were marked.

        """
        if gameboard is None:
            gameboard = self.gameboard
        if player_mark is None:
            player_mark = self.players[0].mark

        if coordinate not in self.get_available_moves(gameboard, player_mark):
            logger.warning('You cannot go here!')
            return 0

        grid: str = gameboard.grid_as_string
        # the cache was filled in :meth:`.get_available_moves` method.
        enemy_coordinates: Directions = tuple(
            self._available_moves_cache[grid, player_mark][coordinate]
        )
        score: int = gameboard.place_mark(coordinate, player_mark)

        for enemy_coordinate in enemy_coordinates:
            score += gameboard.place_mark(
                coordinate=enemy_coordinate, mark=player_mark, force=True,
            )
        if gameboard is self.gameboard:
            # if gameboard is not fake
            del self._available_moves_cache[grid, player_mark]
        return score

    def get_available_moves(
        self,
        gameboard: Optional[SquareGameboard] = None,
        player_mark: Optional[PlayerMark] = None,
    ) -> Coordinates:
        """Determine available cells, save their coordinates to cache.

        :param gameboard: Optional.  If undefined, use
            :attr:`.gameboard`.
        :param player_mark: Optional.  If undefined, use mark of player
            with index ``0`` in :attr:`.players` (current player mark).

        :returns: All coordinates of the given ``gameboard`` where
            player with the ``player_mark`` can move.

        """
        if gameboard is None:
            gameboard = self.gameboard
        if player_mark is None:
            player_mark = self.players[0].mark

        grid: str = gameboard.grid_as_string

        if (grid, player_mark) not in self._available_moves_cache:
            self._available_moves_cache[grid, player_mark] = defaultdict(list)
            enemy_mark: PlayerMark = self.get_enemy_mark(player_mark)
            for cell in filter(
                lambda x: x.mark != enemy_mark, gameboard.cells
            ):
                self._fill_available_moves_cache(
                    gameboard=gameboard,
                    start_coordinate=cell.coordinate,
                    cache_key=(grid, player_mark),
                    player_mark=player_mark,
                    enemy_mark=enemy_mark,
                    reverse=False if cell.mark == EMPTY else True,
                )
        return tuple(self._available_moves_cache[grid, player_mark])

    def get_score(
        self, gameboard: SquareGameboard, player_mark: PlayerMark
    ) -> int:
        """Return difference in number of marks between players.

        :param gameboard: The gameboard relative to which the score of
            the game will be calculated.
        :param player_mark: The player mark relative to whom the score
            of the game will be calculated.

        :returns: Score of the game as the difference in the scoring of
            ``player`` marks and mark of the another player.

        """
        return gameboard.count(player_mark) - gameboard.count(
            self.get_enemy_mark(player_mark)
        )

    def _get_winners(
        self, *, gameboard: SquareGameboard
    ) -> Tuple[PlayerMark, ...]:
        """Return players who have the maximum count of marks.

        :param gameboard: The gameboard relative to which the winner(s)
            will be determined.

        :returns: Tuple with player marks who were determined as
            winners.

        """
        first_player_score: int = gameboard.count(self.players[0].mark)
        second_player_score: int = gameboard.count(self.players[1].mark)
        if first_player_score > second_player_score:
            return (self.players[0].mark,)
        elif first_player_score < second_player_score:
            return (self.players[1].mark,)
        return (self.players[0].mark, self.players[1].mark)

    def _fill_available_moves_cache(
        self,
        gameboard: SquareGameboard,
        start_coordinate: Coordinate,
        cache_key: Tuple[str, PlayerMark],
        player_mark: PlayerMark,
        enemy_mark: PlayerMark,
        reverse: bool = False,
    ) -> None:
        """Fill cache with available moves.

        :param gameboard:  The gameboard that will be checked.
        :param start_coordinate:  The coordinate relative to which the
            directions will be checked.
        :param cache_key:  The key of cache.
        :param player_mark:  The current player mark.
        :param enemy_mark:  The mark to which all cells should have in
            checked directions between ``start_coordinate`` and cell
            with ``player_mark`` (when ``reverse=False``) or ``EMPTY``
            (when ``reverse=True``).
        :param reverse: Optional.  ``False`` by default. If ``False``,
            ``start_coordinate`` will be saved as an available move,
            otherwise an empty cell in each direction where the rules of
            the Reversi game are executed will be saved as an available
            move.

        """
        for offset in gameboard.get_offsets(start_coordinate):
            if gameboard[offset.coordinate].mark == enemy_mark:
                enemy_coordinates: List[Coordinate] = [offset.coordinate]
                next_coordinate, next_mark = gameboard.get_offset_cell(
                    coordinate=offset.coordinate, direction=offset.direction,
                )
                while next_mark == enemy_mark:
                    enemy_coordinates.append(next_coordinate)
                    next_coordinate, next_mark = gameboard.get_offset_cell(
                        coordinate=next_coordinate, direction=offset.direction,
                    )
                if reverse and next_mark == EMPTY:
                    self._available_moves_cache[cache_key][
                        next_coordinate
                    ].extend(enemy_coordinates)
                elif not reverse and next_mark == player_mark:
                    self._available_moves_cache[cache_key][
                        start_coordinate
                    ].extend(enemy_coordinates)



from __future__ import annotations

from collections import deque
import sys
from typing import TYPE_CHECKING

from ap_games.ap_types import EMPTY
from ap_games.ap_types import GameStatus
from ap_games.ap_types import O_MARK
from ap_games.ap_types import X_MARK
from ap_games.gameboard.gameboard import SquareGameboard
from ap_games.log import logger
from ap_games.player.ai_player import AIPlayer
from ap_games.player.human_player import HumanPlayer

if TYPE_CHECKING:
    from typing import ClassVar
    from typing import DefaultDict
    from typing import Deque
    from typing import Dict
    from typing import List
    from typing import Optional
    from typing import Tuple

    from ap_games.ap_types import Coordinate
    from ap_games.ap_types import Coordinates
    from ap_games.ap_types import OptionalPlayerTypes
    from ap_games.ap_types import PlayerMark
    from ap_games.ap_types import SupportedPlayers
    from ap_games.player.player import Player

__all__ = ('TwoPlayerBoardGame',)


class TwoPlayerBoardGame:
    """TwoPlayerBoardGame class specifies the public methods of game.

    Then concrete classes providing the standard game implementations.

    .. note::

        The base class also provide default implementations of some
        methods in order to help implementation of concrete game class.

    :param grid: String contains symbols from set
        :attr:`.TwoPlayerBoardGame.marks` and symbols '_' or ' ' mean an
        empty cell.
    :param player_types: A tuple of strings with two elements from
        :attr:`.TwoPlayerBoardGame.supported_players.keys` which
        determine the types of players.

    :ivar status: This is current status of the game.  ``False`` if game
        can't be continued.
    :ivar gameboard: The gameboard as instance of
        :class:`SquareGameboard`.
    :ivar players: The queue with players.  Player is an instance of
        :class:`Player`.  Player with index ``0`` is a current player.
    :ivar _available_moves_cache: Cache with available moves as dict.
        Where key of dict is a tuple with two fields:

        * ``grid`` of the gameboard before considered move;
        * ``mark`` of player, who makes move on this turn.

        Value of dict is a sub-dict.  Where the keys are the coordinates
            of all available moves, and the sub-dict value are the tuple
            of coordinates of the cells marked by the enemy, which
            should be marked in the considered move.

    """

    marks: ClassVar[Tuple[PlayerMark, PlayerMark]] = (X_MARK, O_MARK)

    default_grid: ClassVar[str] = EMPTY * 9
    grid_axis: ClassVar[bool] = True
    grid_gap: ClassVar[str] = ' '

    supported_players: ClassVar[SupportedPlayers] = {
        'user': HumanPlayer,
        'easy': AIPlayer,
        'medium': AIPlayer,
        'hard': AIPlayer,
        'nightmare': AIPlayer,
    }

    rules: ClassVar[str] = ''
    priority_coordinates: ClassVar[Dict[Coordinate, int]] = {}

    def __init__(
        self,
        *,
        grid: str = '',
        player_types: Tuple[str, str] = ('user', 'user'),
    ):
        if not grid:
            grid = self.default_grid

        if len(player_types) != 2:
            raise ValueError('The number of players should be 2!')

        self.players: Deque[Player] = deque()
        self._add_players(player_types=player_types)

        grid_without_underscore = grid.replace('_', EMPTY)
        if not set(grid_without_underscore).issubset({*self.marks, EMPTY}):
            raise ValueError(
                'Gameboard must contain only " ", "_" and symbols '
                f'from {self.marks}.'
            )

        self.status: GameStatus = GameStatus(
            active=True, message='', must_skip=False
        )
        self.gameboard: SquareGameboard = SquareGameboard(
            grid=grid_without_underscore,
            gap=self.grid_gap,
            axis=self.grid_axis,
        )

        self._available_moves_cache: Dict[
            Tuple[str, PlayerMark], DefaultDict[Coordinate, List[Coordinate]]
        ] = {}

    def play(self) -> None:
        """Start new game."""
        logger.info(self.gameboard)
        self.status = self.get_status()
        while self.status.active:
            coordinate: Coordinate = self.players[0].move()
            if (
                coordinate != self.gameboard.undefined_coordinate
                and self.place_mark(coordinate)
            ):
                logger.info(self.gameboard)
                self.players.rotate(1)
                self.status = self.get_status()
                if self.status.message:
                    logger.info(self.status.message)
                if self.status.must_skip:
                    self.players.rotate(1)
                    self.status = self.status._replace(active=True)
        self._available_moves_cache.clear()

    def get_status(
        self,
        gameboard: Optional[SquareGameboard] = None,
        player_mark: Optional[PlayerMark] = None,
    ) -> GameStatus:
        """Return game status calculated in accordance with game rules.

        .. warning::

            Must be overridden by subclasses if there is a more complex
            rule for calculating game status.

        .. note::

            If there is no available moves for the ``player`` the method
            must return ``GameStatus.active == False`` and
            ``GameStatus.must_skip == True``.

        :param gameboard: Optional.  If undefined, use
            :attr:`.TwoPlayerBoardGame.gameboard`.
        :param player_mark: Optional.  If undefined, use mark of player
            with index ``0`` in :attr:`.players` (current player mark).

        :returns: Game status as the instance of namedtuple
            ``GameStatus`` with three fields: ``active``, ``message``
            and ``must_skip``.   ``GameStatus.active == False`` if game
            cannot be continued.

        """
        if gameboard is None:
            gameboard = self.gameboard
        if player_mark is None:
            player_mark = self.players[0].mark

        if self.get_available_moves(gameboard, player_mark):
            return GameStatus(active=True, message='', must_skip=False)
        return GameStatus(active=False, message='', must_skip=False)

    def place_mark(
        self,
        coordinate: Coordinate,
        player_mark: Optional[PlayerMark] = None,
        gameboard: Optional[SquareGameboard] = None,
    ) -> int:
        """Change the mark of the cell with coordinate on the gameboard.

        :param coordinate: coordinate of cell which player mark.
        :param player_mark: Optional.  If undefined, use mark of player
            with index ``0`` in :attr:`.players` (current player mark).
        :param gameboard: Optional.  If undefined, use
            :attr:`.TwoPlayerBoardGame.gameboard`.

        This method should be overridden by subclasses if there is a
        more complex rule for marking cell(s) in ``gameboard``.

        :returns: Score as count of marked cells.  Return ``0`` if no
            cells were marked.

        """
        if gameboard is None:
            gameboard = self.gameboard
        if player_mark is None:
            player_mark = self.players[0].mark

        if coordinate not in self.get_available_moves(gameboard):
            logger.warning('You cannot go here!')
            return 0
        return gameboard.place_mark(coordinate, player_mark)

    def get_available_moves(
        self,
        gameboard: Optional[SquareGameboard] = None,
        player_mark: Optional[PlayerMark] = None,
    ) -> Coordinates:
        """Return coordinates of all ``EMPTY`` cells on the gameboard.

        This method should be overridden by subclasses if there is a
        more complex rule for determining which cell is available.

        :param gameboard: Optional.  If undefined, use
            :attr:`.TwoPlayerBoardGame.gameboard`.
        :param player_mark: Optional.  If undefined, use mark of player
            with index ``0`` in :attr:`.players` (current player mark).

        :returns: All coordinates of the given ``gameboard`` where
            player with the ``player_mark`` can move.

        """
        if gameboard is None:
            gameboard = self.gameboard
        return gameboard.available_moves

    def get_score(
        self, gameboard: SquareGameboard, player_mark: PlayerMark
    ) -> int:
        """Return the score relative to the given gameboard and player.

        :param gameboard: The gameboard relative to which the score of
            the game will be calculated.
        :param player_mark: The player mark relative to whom the score
            of the game will be calculated.

        :returns: Score of the game from ``-1`` to ``1``, where ``-1``
            corresponds to a loss of ``player``; ``0`` corresponds to
            draw and ``1`` corresponds to win of ``player``.

        """
        winners: Tuple[PlayerMark, ...] = self._get_winners(
            gameboard=gameboard
        )
        if len(winners) == 1:
            if player_mark in winners:
                return 1
            else:
                return -1
        else:  # len(winners) != 1
            return 0

    def get_next_player(self, current_player: Player) -> Player:
        """Return ``player`` who is next to ``current_player``.

        :param current_player: The player relative to whom the another
            player will be calculated.

        :returns: Player who will be move next after ``current_player``.

        """
        return (
            self.players[1]
            if current_player is self.players[0]
            else self.players[0]
        )

    def cli(self, player_types: OptionalPlayerTypes = ()) -> None:
        """Command line interface of the game.

        :param player_types: A tuple of strings with two elements from
            :attr:`.TwoPlayerBoardGame.supported_players.keys` which
            determine the types of players.

        """
        if player_types and len(player_types) == 2:
            command: str = f'start {player_types[0]} {player_types[1]}'
        else:
            logger.info(
                'Type "start user1_type user2_type" to run the game, '
                'where "user1_type" and "user2_type" is one of the '
                f'supported values: {", ".join(self.supported_players)}; '
                'Type "rules" to get game rules or type "exit" to '
                'exit from the game.\nInput command: '
            )
            command = sys.stdin.readline().strip()
        logger.debug(f'{command=}')
        while command != 'exit':
            parameters = command.split()
            if (
                len(parameters) == 3
                and parameters[0] == 'start'
                and parameters[1] in self.supported_players
                and parameters[2] in self.supported_players
            ):
                self._add_players(player_types=(parameters[1], parameters[2]))
                self.gameboard = SquareGameboard(
                    grid=self.default_grid,
                    gap=self.grid_gap,
                    axis=self.grid_axis,
                )
                self.play()
            elif command == 'rules':
                logger.info(self.rules)
            else:
                logger.warning('Bad parameters!')
            logger.info('\nInput command: ')
            command = sys.stdin.readline().strip()
            logger.debug(f'{command=}')

    def _add_players(self, *, player_types: Tuple[str, str]) -> None:
        """Create instances of `Player` and add them to :attr:`.players`.

        :param player_types: A tuple of strings with two elements from
            :attr:`.TwoPlayerBoardGame.supported_players`.keys() which
            determine the types of players.

        """
        self.players.clear()
        for num, player_type in enumerate(player_types):
            self.players.append(
                self.supported_players[player_type](
                    player_type, mark=self.marks[num], game=self  # noqa: T484
                )
            )

    def _get_winners(
        self, *, gameboard: SquareGameboard
    ) -> Tuple[PlayerMark, ...]:
        """Return a tuple of :class:`Player` instances defined as winner(s).

        .. warning::

            This method must be overridden by subclasses.

        :param gameboard: The gameboard relative to which the winner(s)
            will be determined.

        :returns: Tuple with player marks who were determined as
            winners.

        """
        return ()

    def _rotate_players(self) -> None:
        """Move player with the least number of mark to the front of queue."""
        while self.gameboard.count(
            self.players[0].mark
        ) > self.gameboard.count(self.players[1].mark):
            self.players.rotate(1)

    @staticmethod
    def get_enemy_mark(player_mark: PlayerMark) -> PlayerMark:
        """Return enemy mark relatively to ``player_mark``.

        :param player_mark: The player's mark, relative to which the
            enemy player's mark will be determined.

        :returns: The opponent's player's mark.

        """
        return X_MARK if player_mark == O_MARK else O_MARK



from __future__ import annotations

from functools import cached_property
import random
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Tuple

    from ap_games.ap_types import Coordinate
    from ap_games.ap_types import PlayerMark
    from ap_games.game.game_base import TwoPlayerBoardGame

__all__ = ('Player',)


class Player:
    """Class introduces the player in a board game."""

    def __init__(
        self, type_: str, /, *, mark: PlayerMark, game: TwoPlayerBoardGame
    ):
        self._type: str = type_
        self._game: TwoPlayerBoardGame = game
        self._mark: PlayerMark = mark

    def __str__(self) -> str:
        """Return mark."""
        return self._mark

    @cached_property
    def type_(self) -> str:
        """Return type."""
        return self._type

    @cached_property
    def game(self) -> TwoPlayerBoardGame:
        """Return game as instance of :class:`TwoPlayerBoardGame`."""
        return self._game

    @cached_property
    def mark(self) -> PlayerMark:
        """Return mark."""
        return self._mark

    def move(self) -> Coordinate:
        """Return the randomly selected coordinate.

        .. note::

            This method should be overridden by subclasses if there is a
            more complex rule for determining coordinate of move.

        :returns: User-selected cell coordinate.

        """
        return self._random_coordinate()

    def _random_coordinate(self) -> Coordinate:
        """Return randomly selected coordinate on the gameboard.

        :returns: One randomly selected coordinate if there are
            available coordinates, else ``undefined_coordinate``.

        """
        available_moves: Tuple[
            Coordinate, ...
        ] = self.game.get_available_moves()
        return (
            random.choice(available_moves)
            if available_moves
            else self.game.gameboard.undefined_coordinate
        )



import logging
import os
import pathlib
import sys

__all__ = ('logger',)

BASE_DIR = pathlib.Path(__file__).parent.parent.parent.resolve(strict=True)
LOG_LEVEL = os.environ.get('AP_GAMES_LOGLEVEL', 'ERROR')
LOG_FILE = os.environ.get('AP_GAMES_LOGFILE', f'{BASE_DIR}/ap_games.log')


file_handler = logging.FileHandler(LOG_FILE)
console_handler = logging.StreamHandler(sys.stdout)
handlers = [file_handler, console_handler]
file_handler.setLevel(logging.WARNING)
console_handler.setLevel(logging.DEBUG)
logging.basicConfig(
    format='%(message)s', level=logging.INFO, handlers=handlers,
)

logger = logging.getLogger()



from __future__ import annotations

from configparser import ConfigParser
from importlib import resources
import random
import sys
from typing import NamedTuple
from typing import TYPE_CHECKING

from ap_games.game.reversi import Reversi
from ap_games.game.tictactoe import TicTacToe
from ap_games.log import logger

__all__ = ('main',)


if sys.version_info < (3, 8):
    raise RuntimeError('This package requires Python 3.8+!')


if TYPE_CHECKING:
    from typing import Dict
    from typing import Tuple
    from typing import Type

    from ap_games.ap_types import OptionalPlayerTypes
    from ap_games.game.game_base import TwoPlayerBoardGame


TEST_MODE: bool = False


class Game(NamedTuple):
    """Game(name: str, game_class: Type[TwoPlayerBoardGame])."""

    name: str
    game_class: Type[TwoPlayerBoardGame]


supported_games: Dict[str, Game] = {
    '1': Game(name='Tic-Tac-Toe', game_class=TicTacToe),
    '2': Game(name='Reversi', game_class=Reversi),
}


def main() -> None:
    """Aks user about desired game and run it."""
    choice: str
    player_types: OptionalPlayerTypes

    read_config()
    if TEST_MODE:
        run_test_mode_and_exit()
    choice, player_types = read_argv()
    games: str = ";\n\t".join(
        f'{num} - {game.name}' for num, game in supported_games.items()
    )
    message: str = (
        f'Please choose the game:\n\t{games}.\n'
        'Print "exit" to exit the program.\n\nInput command: '
    )
    while choice != 'exit':
        if choice in supported_games:
            logger.debug(f'{choice=}')
            game: TwoPlayerBoardGame = supported_games[choice].game_class()
            game.cli(player_types=player_types)
        logger.info(message)
        choice = sys.stdin.readline().strip()


def read_config() -> None:
    """Read the log level from the config.ini and set it."""
    cfg = ConfigParser()
    cfg.read_string(
        resources.read_text(package='ap_games', resource='config.ini')
    )
    log_level: str = cfg.get('ap-games', 'log_level').upper()
    logger.setLevel(log_level if log_level == 'DEBUG' else 'INFO')
    global TEST_MODE
    TEST_MODE = cfg.getboolean('ap-games', 'test_mode')


def run_test_mode_and_exit() -> None:
    """Run the predefined configuration if ``TEST_MODE=True`` and exit."""
    random.seed(42)
    logger.debug(f'{TEST_MODE=}')
    game: TwoPlayerBoardGame = Reversi(player_types=('medium', 'hard'))
    game.play()
    game = TicTacToe(player_types=('easy', 'hard'))
    game.play()
    sys.exit()


def read_argv() -> Tuple[str, OptionalPlayerTypes]:
    """Read command-line arguments and return them.

    :returns: Two-element tuple, where:

        * ``game_num`` - a value from ``supported_games.keys()`` or
            empty string;
        * ``player_types`` - a two-element tuple, where each element is
          a player-type string or empty tuple.

    """
    sys.argv.pop(0)
    game_num: str = ''
    player_types: OptionalPlayerTypes = ()
    if len(sys.argv) >= 1:
        game_num = sys.argv[0]
        game_num = game_num.title()
        for num, game in supported_games.items():
            game_num = game_num.replace(game.name, num)
        if len(sys.argv) >= 3:
            player_types = (sys.argv[1], sys.argv[2])
    return (game_num, player_types)


if __name__ == '__main__':
    main()



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from platform import platform
from typing import cast
from typing import TYPE_CHECKING

from ap_games.ap_types import Cell
from ap_games.ap_types import Coordinate
from ap_games.ap_types import EMPTY
from ap_games.ap_types import Mark
from ap_games.ap_types import O_MARK
from ap_games.ap_types import Offset
from ap_games.ap_types import X_MARK
from ap_games.log import logger

if TYPE_CHECKING:
    from typing import Any
    from typing import ClassVar
    from typing import Dict
    from typing import Final
    from typing import List
    from typing import Tuple

    from ap_games.ap_types import Coordinates
    from ap_games.ap_types import Directions
    from ap_games.ap_types import PlayerMark
    from ap_games.ap_types import Side
    from ap_games.ap_types import Size

__all__ = ('SquareGameboard',)


@dataclass(frozen=True)
class _Colors:
    header: Final[str] = '\033[35m'
    blue: Final[str] = '\033[34m'
    green: Final[str] = '\033[32m'
    yellow: Final[str] = '\033[33m'
    turquoise: Final[str] = '\033[36m'
    end: Final[str] = '\033[0m'
    bold: Final[str] = '\033[1m'


class _GameboardRegistry:
    """GameboardRegistry stores basic mapping of SquareGameboard class.

    :param size: The size of SquareGameboard that used in computation
        mappings.

    :ivar index_to_coordinate: Store the mapping of the cell index to
        the coordinate of this cell.  All grid cells are indexed from
        left to right from top to bottom.

        .. seealso::

            :meth:`._fill_index_to_coordinate`

    :ivar offsets: Store the mapping of the grid coordinates to its
        offsets.  Where each offset is a tuple with two fields:

        * ``coordinate`` of the neighboring cell;
        * ``direction`` of the neighboring cell as a coordinate
          with ``-1<=x<=1`` and ``-1<=y<=1`` relatively to the base
          coordinate.

        .. seealso::

            :meth:`._fill_offsets`

    :ivar all_coordinates: Store all possible coordinates of the
        :class:`SquareGameboard` with the given ``size``.

    """

    _directions: Final[ClassVar[Directions]] = (
        Coordinate(0, 1),  # top
        Coordinate(1, 1),  # right-top
        Coordinate(1, 0),  # right and so on
        Coordinate(1, -1),
    )

    def __init__(self, *, size: int) -> None:
        if (size < 2) or (size > 9):
            raise ValueError(
                'The size of the gameboard must be between 2 and 9!'
            )
        self.size = size
        self.index_to_coordinate: Dict[int, Coordinate] = {}
        self.offsets: Dict[Coordinate, Tuple[Offset, ...]] = {}
        self._fill_index_to_coordinate()
        self.all_coordinates: Final[Coordinates] = tuple(
            self.index_to_coordinate.values()
        )
        self._fill_offsets()

    def _fill_index_to_coordinate(self) -> None:
        """Convert index of cell into coordinate of cell.

        Where an example for a 3x3 grid::

            0 1 2         (1, 3) (2, 3) (3, 3)
            3 4 5   ==>   (1, 2) (2, 2) (3, 2)
            6 7 8         (1, 1) (2, 1) (3, 1)

        """
        for index in range(self.size ** 2):
            a, b = divmod(index, self.size)
            column: int = b + 1
            row: int = self.size - a
            self.index_to_coordinate[index] = Coordinate(column, row)

    def _fill_offsets(self) -> None:
        """Fill up :attr:`.offsets` for all coordinates of gameboard."""
        for coordinate in self.all_coordinates:
            offsets: List[Offset] = []
            for shift in self._directions:
                offset_coordinate = Coordinate(
                    x=coordinate.x + shift.x, y=coordinate.y + shift.y
                )
                if offset_coordinate in self.all_coordinates:
                    offsets.append(
                        Offset(coordinate=offset_coordinate, direction=shift)
                    )
            self.offsets[coordinate] = tuple(offsets)


class SquareGameboard:
    """Implementation square game board with size from 2 to 9.

    :param grid: The grid or board, represented as a string, where
        each character is mapped to a cell left to right top to bottom.
    :param gap: ``' '`` by default.  Defines the gap that will be
        printed between cells in a row.
    :param axis: ``False`` by default.  If ``True`` print axis.
    :param colorized: ``True`` by default.  Determines whether to use
        colors from :class:`_Colors` when gameboard printed.

        .. warning::

            This attribute is always equal ``False`` on non-Linux
            platforms.

    :param indent:  The indentation that will be added to the front of
        grid when printing the gameboard.

    :ivar _size: The size of gameboard from 2 to 9.
    :ivar _cells_dict: Dict with cells of the gameboard.  Where key is
        a tuple with two coordinates of the corresponding cell.  And the
        value is an instance of :class:`Cell`.

    :ivar _colors_dict: Dict with colors of each cell of the gameboard.
        Where key is a tuple with two coordinates of the corresponding
        cell.  And the value is a value of :class:`_Colors`.
    :ivar _grid_cache: Save the grid of the gameboard as a string with
        cell labels. This cache is cleared in :meth:`.place_mark`.

        .. note::

            This attribute adds 25% processing speed.

    """

    undefined_coordinate: Final[ClassVar[Coordinate]] = Coordinate(x=0, y=0)
    undefined_cell: Final[ClassVar[Cell]] = Cell(
        coordinate=undefined_coordinate, mark=''
    )

    mark_colors: ClassVar[Dict[str, str]] = {
        X_MARK: _Colors.blue,
        O_MARK: _Colors.green,
        EMPTY: _Colors.header,
    }

    default_grid: ClassVar[str] = EMPTY * 9

    _registries: Dict[Size, _GameboardRegistry] = {}

    def __new__(cls, **kwargs: Any) -> Any:
        """Create instance and set :attr:`._registries[size]` if necessary."""
        _safety: bool = kwargs.get('_safety', True)
        if _safety:
            grid: str = kwargs.get('grid', cls.default_grid)
            size: int = int(len(grid) ** (1 / 2))
            if size ** 2 != len(grid):
                raise ValueError(
                    f'The gameboard must be square ({size}^2 != {len(grid)})!'
                )
            if size not in cls._registries:
                cls._registries[size] = _GameboardRegistry(size=size)
        return super().__new__(cls)

    def __init__(
        self,
        grid: str = default_grid,
        *,
        gap: str = ' ',
        axis: bool = False,
        colorized: bool = True,
        indent: str = '',
        _safety: bool = True,
    ) -> None:

        self.colorized: bool = colorized if _safety and platform().startswith(
            'Linux'
        ) else False
        self.indent: str = indent
        size: int = int(len(grid) ** (1 / 2))
        self.registry: _GameboardRegistry = self._registries[size]
        self._size: Final[int] = size
        self._gap: Final[str] = gap
        self._axis: Final[bool] = axis

        self._cells_dict: Dict[Tuple[int, int], Cell] = {}
        self._colors_dict: Dict[Tuple[int, int], str] = {}
        self._grid_cache: str = ''
        if _safety:
            for index, mark in enumerate(grid):
                coordinate = self.registry.index_to_coordinate[index]
                mark = cast(Mark, mark)
                self._cells_dict[coordinate] = Cell(coordinate, mark)
            self._default_paint()

        self._horizontal_border: str = (
            ('  ' if self._axis else '')
            + '-' * (self._size + len(self._gap) * (self._size + 1) + 2)
        )
        _column_nums: str = f'{self._gap}'.join(
            map(str, range(1, self._size + 1))
        )
        self._column_axis: str = (
            f'\n{self.indent}   {self._gap}{_column_nums}{self._gap}'
            if self._axis
            else ''
        )

    def __str__(self) -> str:
        """Return SquareGameboard as a grid drawn with ASCII characters.

        See example with ``size=4``, ``gap=' '`` and ``axis=True``::

              -----------
            4 |         |
            3 |   X O   |
            2 |   O X   |
            1 |         |
              -----------
                1 2 3 4

        """
        grid: str = '\n'.join(
            self.indent
            + (f'{num} ' if self._axis else '')
            + f'|{self._gap}'
            + f'{self._gap}'.join(
                (self._colors_dict[cell.coordinate] if self.colorized else '')
                + cell.mark
                + (_Colors.end if self.colorized else '')
                for cell in row
            )
            + f'{self._gap}|'
            for num, row in reversed(tuple(enumerate(self.rows, 1)))
        )

        return (
            f'{self.indent}{self._horizontal_border}\n'
            f'{grid}\n{self.indent}{self._horizontal_border}'
            f'{self._column_axis}'
        )

    def __getitem__(self, coordinate: Coordinate) -> Cell:
        """Return Cell by coordinate."""
        return self._cells_dict[coordinate]

    @cached_property
    def size(self) -> int:
        """Return size of gameboard."""
        return self._size

    @property
    def grid_as_string(self) -> str:
        """Return all marks of the gameboard as a string.

        Example::

            -------
            | 0 1 |
            | 2 3 |   ===>   '0123'
            -------

        """
        if not self._grid_cache:
            self._grid_cache = ''.join(
                cell.mark for cell in self._cells_dict.values()
            )
        return self._grid_cache

    @property
    def columns(self) -> Tuple[Side, ...]:
        """Return all columns of gameboard as a tuple."""
        return tuple(
            tuple(
                filter(
                    lambda cell: cell.coordinate.x == column,
                    self._cells_dict.values(),
                )
            )
            for column in range(1, self._size + 1)
        )

    @property
    def rows(self) -> Tuple[Side, ...]:
        """Return all rows of gameboard as a tuple."""
        return tuple(
            tuple(
                filter(
                    lambda cell: cell.coordinate.y == row,
                    self._cells_dict.values(),
                )
            )
            for row in range(1, self._size + 1)
        )

    @property
    def diagonals(self) -> Tuple[Side, Side]:
        """Return main and reverse diagonals as tuples of cell."""
        main_diagonal: Side = tuple(
            self._cells_dict[num, self._size - num + 1]
            for num in range(1, self._size + 1)
        )
        reverse_diagonal: Side = tuple(
            self._cells_dict[num, num] for num in range(1, self._size + 1)
        )
        return main_diagonal, reverse_diagonal

    @property
    def all_sides(self) -> Tuple[Side, ...]:
        """Return all rows, columns and diagonals as tuple of all sides.

        Where each side is a tuple of cells of the corresponding side.

        """
        return self.rows + self.columns + self.diagonals

    @property
    def cells(self) -> Tuple[Cell, ...]:
        """Return all cells of the gameboard.

        Where each cell is a namedtuple with two fields:

        * ``coordinate`` - as namedtuple ``Coordinate`` with two
          ``int`` fields ``x`` and ``y``;
        * ``mark`` of cell as a one-character string.

        """
        return tuple(self._cells_dict.values())

    @property
    def available_moves(self) -> Coordinates:
        """Return coordinates of all ``EMPTY`` cells."""
        return tuple(
            cell.coordinate
            for cell in self._cells_dict.values()
            if cell.mark == EMPTY
        )

    def place_mark(
        self, coordinate: Coordinate, mark: PlayerMark, *, force: bool = False,
    ) -> int:
        """Mark cell of the gameboard with the ``coordinate``.

        :param coordinate:  Position of cell as instance of namedtuple
            Coordinate(x, y).
        :param mark:  New mark.  It will be set if ``force=True`` or
            cell with ``coordinate`` is **empty** (``EMPTY``).
        :param force:  ``False`` by default.  If ``True`` it doesn't
            matter if cell is **empty** or not.

        :returns:  Count of marked cell with ``mark``.

        """
        if (
            force
            or self._cells_dict.get(coordinate, self.undefined_cell).mark
            == EMPTY
        ):
            self._cells_dict[coordinate] = Cell(coordinate, mark)
            if self.colorized:
                if force:
                    self._colors_dict[coordinate] = _Colors.turquoise
                else:
                    self._default_paint()
                    self._colors_dict[
                        coordinate
                    ] = f'{_Colors.bold}{_Colors.turquoise}'
            self._grid_cache = ''
            return 1
        logger.warning('This cell is occupied! Choose another one!')
        return 0

    def get_offset_cell(
        self, *, coordinate: Coordinate, direction: Coordinate
    ) -> Cell:
        """Return cell by coordinate calculated as algebraic sum of parameters.

        :param coordinate: Coordinate of starting cell;
        :param direction: Shift as a coordinate with ``-1<=x<=1`` and
            ``-1<=y<=1``.

        :returns: adjacent cell as instance of :class:`Cell` if cell
            exist on the gameboard else :attr:`.undefined_cell`.

        """
        return self._cells_dict.get(
            (coordinate.x + direction.x, coordinate.y + direction.y,),
            self.undefined_cell,
        )

    def get_offsets(self, coordinate: Coordinate) -> Tuple[Offset, ...]:
        """Return the offsets of the given coordinate.

        :param coordinate:  The coordinate from which the offsets
            will be returned.

        :raises ValueError: If ``coordinate`` out of range.

        :returns: Tuple of offsets, where each offset is an instance of
            namedtuple :class:`Offset`.

        """
        if coordinate not in self.registry.all_coordinates:
            raise ValueError(f'The {coordinate} out of range!')
        return self.registry.offsets[coordinate]

    def count(self, mark: str) -> int:
        """Return the number of occurrences of ``mark`` on the gameboard.

        :param mark:  The mark relative to which the cell will be
            matched by.

        """
        return self.grid_as_string.count(mark)

    def copy(self, **kwargs: Any) -> SquareGameboard:
        """Return copy of current gameboard with exactly the same grid.

        :param kwargs: Arguments that will be override default
            arguments.

        :returns: The new instance with the same values of attributes.

        """
        sg: SquareGameboard = SquareGameboard(
            grid=self.grid_as_string,
            gap=self._gap,
            axis=self._axis,
            _safety=False,
            **kwargs,
        )
        sg._cells_dict = dict(self._cells_dict)
        sg._grid_cache = self._grid_cache
        return sg

    def _default_paint(self) -> None:
        self._colors_dict = {
            coordinate: self.mark_colors[self._cells_dict[coordinate].mark]
            for coordinate in self._cells_dict
        }



from setuptools import setup  # type: ignore

setup(
    extras_require=dict(
        test=['coverage==5.2.1', 'pytest==6.0.1', 'pytest-cov==2.10.1'],
        dev=[
            'black==19.10b0',
            'bump2version==1.0.0',
            'check-manifest==0.42',
            'darglint==1.5.2',
            'flake8==3.8.3',
            'flake8-bugbear==20.1.4',
            'flake8-builtins==1.5.3',
            'flake8-docstrings==1.5.0',
            'flake8-import-order==0.18.1',
            'flake8-mypy==17.8.0',
            'flake8-pytest-style==1.2.3',
            'flake8-rst-docstrings==0.0.13',
            'flake8-typing-imports==1.9.0',
            'keyring==21.3.0',
            'mypy==0.782',
            'mypy-extensions==0.4.3',
            'pre-commit==2.6.0',
            'pre-commit-hooks==3.2.0',
            'pydocstyle==5.0.2',
            'pyupgrade==2.7.2',
            'tox==3.19.0',
            'twine==3.2.0',
        ],
    ),
)



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from ap_games.ap_types import Coordinate
from ap_games.log import logger
from ap_games.player.player import Player

if TYPE_CHECKING:
    from typing import List

__all__ = ('HumanPlayer',)


class HumanPlayer(Player):
    """HumanPlayer in a game with interaction through the CLI."""

    def move(self) -> Coordinate:
        """Read coordinate of the next move from the input and return it.

        :returns: Return :attr:`.SquareGameboard.undefined_coordinate`
            if the coordinate is incorrect.

        """
        logger.info(f'Enter the coordinate [{self._mark}]: ')
        input_list: List[str] = sys.stdin.readline().strip().split()
        logger.debug(f'{input_list=}')
        if len(input_list) >= 2:
            column, row = input_list[:2]
            if column.isdigit() and row.isdigit():
                return Coordinate(int(column), int(row))
        logger.warning('You should enter two numbers!')
        return self.game.gameboard.undefined_coordinate



from __future__ import annotations

from typing import TYPE_CHECKING

from ap_games.ap_types import GameStatus
from ap_games.game.game_base import TwoPlayerBoardGame

if TYPE_CHECKING:
    from typing import Any
    from typing import ClassVar
    from typing import List
    from typing import Optional
    from typing import Tuple

    from ap_games.ap_types import PlayerMark
    from ap_games.gameboard.gameboard import SquareGameboard

__all__ = ('TicTacToe',)


class TicTacToe(TwoPlayerBoardGame):
    """Tic-Tac-Toe game supports human user and some types of AI.

    .. seealso::

        :class:`TwoPlayerBoardGame`

    """

    rules: ClassVar[str] = ''.join(
        (
            'Tic-tac-toe, is a paper-and-pencil game for two players, ',
            'X and O, who take turns marking the spaces in a 3×3 grid.\n',
            'The player who succeeds in placing three of their marks in ',
            'a horizontal, vertical, or diagonal row is the winner.',
        )
    )

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._rotate_players()

    def get_status(
        self,
        gameboard: Optional[SquareGameboard] = None,
        player_mark: Optional[PlayerMark] = None,
    ) -> GameStatus:
        """Return difference in number of marks between players.

        :param gameboard: Optional.  If undefined, use
            :attr:`.TwoPlayerBoardGame.gameboard`.
        :param player_mark: This parameter is not used in this
            implementation.

        :returns: Game status as the instance of namedtuple
            ``GameStatus`` with two fields: ``active`` and ``message``.
            ``GameStatus.active == False`` if game cannot be continued.

        """
        if gameboard is None:
            gameboard = self.gameboard

        game_status: GameStatus = GameStatus(
            active=True, message='', must_skip=False
        )
        if (
            abs(
                gameboard.count(self.players[0].mark)
                - gameboard.count(self.players[1].mark)
            )
            > 1
        ):
            game_status = GameStatus(
                active=False, message='Impossible\n', must_skip=False
            )
        else:
            winners: Tuple[PlayerMark, ...] = self._get_winners(
                gameboard=gameboard
            )
            if (not winners) and not self.get_available_moves(gameboard):
                game_status = GameStatus(
                    active=False, message='Draw\n', must_skip=False
                )
            elif len(winners) == 1:
                game_status = GameStatus(
                    active=False,
                    message=f'{winners[0]} wins\n',
                    must_skip=False,
                )
            elif len(winners) > 1:
                game_status = GameStatus(
                    active=False, message='Impossible\n', must_skip=False
                )
        return game_status

    def _get_winners(
        self, *, gameboard: SquareGameboard
    ) -> Tuple[PlayerMark, ...]:
        """Return players who draw solid line.

        If all characters on a 'side' are the same and equal to the
        mark of player from :attr:`.players`, this player is added to
        the set of winners.

        :param gameboard: The gameboard relative to which the winner(s)
            will be determined.

        :returns: Tuple with player marks who were determined as
            winners.

        """
        if gameboard is None:
            gameboard = self.gameboard

        winners: List[PlayerMark] = []
        all_sides_as_strings: List[str] = [
            ''.join(cell.mark for cell in side) for side in gameboard.all_sides
        ]
        first_player_side: str = self.players[0].mark * gameboard.size
        second_player_side: str = self.players[1].mark * gameboard.size
        if first_player_side in all_sides_as_strings:
            winners.append(self.players[0].mark)
        if second_player_side in all_sides_as_strings:
            winners.append(self.players[1].mark)
        return tuple(winners)
